from typing import TYPE_CHECKING, Optional

from PIL import Image

from tarraz.stitcher import Stitcher

if TYPE_CHECKING:
    from tarraz.models import Color, Coordinate


class DisplayStitcher(Stitcher):
    def init(self, width: int, height: int) -> None:
        self.result = Image.new("RGB", (width, height))

    def draw_cell(
        self, coordinate: "Coordinate", size: int, color: "Optional[Color]" = None
    ) -> None:
        pixels = self.result.load()

        for i in range(coordinate.x, coordinate.x + size):
            for j in range(coordinate.y, coordinate.y + size):
                pixels[i - size, j - size] = color.rgb if color else (255, 255, 255, 0)

    def finish(self) -> None:
        self.result.show()

    def save(self, ext: str) -> None:
        return



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent
IMAGE_EXTENSIONS = (".jpeg", ".jpg", ".png", ".webp")
COLORS_EXTENSIONS = (".json",)

SVG_VARIANTS = [
    {
        "name": "key",
        "black_white": False,
        "minor_lines": True,
        "symbols": True,
        "key": True,
    },
    {
        "name": "colored_symbols",
        "black_white": False,
        "minor_lines": True,
        "symbols": True,
        "mid_arrows": True,
        "major_gridlines": True,
    },
    {
        "name": "black_white_symbols",
        "black_white": True,
        "minor_lines": True,
        "symbols": True,
        "mid_arrows": True,
        "major_gridlines": True,
    },
    {
        "name": "colored",
        "black_white": False,
        "minor_lines": False,
        "symbols": False,
        "mid_arrows": False,
        "major_gridlines": False,
    },
]



import json
import logging


class JsonFormatter(logging.Formatter):
    def format(self, record):
        log_record = {
            "timestamp": self.formatTime(record),
            "message": record.getMessage(),
            "severity": record.levelname,
            "module": record.module,
            "line": record.lineno,
        }
        return json.dumps(log_record)


logger = logging.getLogger()
handler = logging.StreamHandler()
handler.setFormatter(JsonFormatter())

logger.addHandler(handler)
logger.setLevel(logging.INFO)



from typing import List, Literal, NamedTuple, TypeVar, Optional


class RGB(NamedTuple):
    red: int
    green: int
    blue: int

    def __str__(self) -> str:
        return f"{self.red}, {self.green}, {self.blue}"

    @property
    def native(self):
        return self.red, self.green, self.blue

    @property
    def css(self):
        return f"rgb({self.red}, {self.green}, {self.blue})"


class Color(object):
    _glyphs_count = 0

    def __init__(self, code: str, rgb: RGB, name: str) -> None:
        self.code = code
        self.rgb = rgb
        self.name = name
        self._glyph: "Optional[int]" = None

    @property
    def glyph(self):
        return self._glyph

    def assign_glyph(self, number: "Optional[int]" = None):
        self._glyph = number if number else Color._glyphs_count + 1
        Color._glyphs_count += 1

    def __str__(self) -> str:
        return self.name

    @classmethod
    def create(
        cls, raw_colors: "List[dict]", extras: "Optional[List[RGB, RGB]]" = None
    ) -> List["Color"]:
        result = []
        processed = set()
        for raw_color in raw_colors:
            rgb = RGB(*raw_color["rgb"])
            if str(rgb) in processed:
                continue

            processed.add(str(rgb))
            result.append(
                cls(
                    code=raw_color["code"],
                    rgb=RGB(*raw_color["rgb"]),
                    name=raw_color["name"],
                )
            )

        extras = extras or []
        for i, extra in enumerate(extras):
            if str(extra) in processed:
                continue

            processed.add(str(extra))
            result.append(
                cls(
                    code=f"{i}",
                    rgb=extra,
                    name=f"extra_{i}",
                )
            )

        return result


class Coordinate(NamedTuple):
    x: int
    y: int


class ImageSize(NamedTuple):
    width: int
    height: int


T = TypeVar("T")

Palette = List["Color"]
ImageRow = List[T]

RGBImageRow = ImageRow[RGB]
PaletteImageRow = ImageRow[int]

RGBImage = List[RGBImageRow]
PaletteImage = List[ImageRow[int]]

StrokeType = Literal["stroke:rgb(20,20,20);stroke-width:1;", "stroke:none;"]
SVGAttributes = tuple[str, str, StrokeType]



from typing import TYPE_CHECKING, List, Union

from PIL import Image

from tarraz.logger import logger
from tarraz.models import RGB, Color, Coordinate, ImageSize
from tarraz.providers import DMCProvider
from tarraz.utils import generate_image, get_neighbours, index_of

if TYPE_CHECKING:
    from tarraz.models import (
        Palette,
        PaletteImage,
        PaletteImageRow,
        RGBImage,
        RGBImageRow,
    )
    from tarraz.providers import ColorProvider


class Tarraz(object):
    def __init__(
        self,
        image_path: str,
        provider: "ColorProvider" = DMCProvider(),
        cleanup: bool = True,
        colors_num: int = 3,
        result_width: int = 1000,
        x_count: int = 50,
    ) -> None:
        self.new_width = result_width

        self._cleanup = cleanup
        self._colors_num = colors_num
        self._image = Image.open(image_path).convert("RGB")
        self._provider = provider
        self._x_count = x_count

    @property
    def pixel_size(self) -> int:
        return self.new_width // int(self._x_count)

    @property
    def size(self) -> "ImageSize":
        return ImageSize(*self._image.size)

    def process(self) -> tuple["PaletteImage", List[Color]]:
        """Create a resized image with the translated colors."""
        logger.info("Processing image started...")

        self._resize_image()
        translated_image = self._translate_image()
        colored_image = generate_image(translated_image)

        # Translate pixels through the palette using the required number of colors.
        self._image = colored_image.convert(
            "P", palette=Image.ADAPTIVE, colors=self._colors_num
        )

        pattern = self._create_pattern()
        colors = self._generate_palette()

        if self._cleanup:
            self._clean_up(pattern)
        else:
            logger.info("Bypassing cleanup job!")

        return pattern, colors

    def _get_pixel(self, x: int, y: int, palette=False) -> Union[int, "RGB"]:
        coordinate = Coordinate(x, y)
        pixel_data = self._image.getpixel(coordinate)

        if palette:
            return pixel_data

        # Ignore alpha value incase we have a png
        return RGB(*pixel_data[:3])

    def _resize_image(self):
        scale = self.new_width / self.size.width
        new_height = int(self.size.height * scale)
        new_size = ImageSize(self.new_width, new_height)

        self._image = self._image.resize(new_size, Image.NEAREST)

    def _translate_image(self) -> "RGBImage":
        """Convert image pixels colors to match the provider colors."""
        logger.info(f"Translating image colors to {self._provider}...")

        colors: "RGBImage" = []
        for y in range(0, self.size.height, self.pixel_size):
            row: "RGBImageRow" = []
            for x in range(0, self.size.width, self.pixel_size):
                pixel = self._get_pixel(x, y)
                matching_color = self._provider.get_matching_color(pixel)
                row.append(matching_color.rgb)
            colors.append(row)

        return colors

    def _create_pattern(self) -> "PaletteImage":
        """Create an image with the information from the new image."""
        logger.info("Generating SVG information...")
        pattern: "PaletteImage" = []

        for y in range(self.size.height):
            row: "PaletteImageRow" = []
            for x in range(self.size.width):
                color_index = self._get_pixel(x, y, palette=True)
                row.append(color_index)
            pattern.append(row)

        return pattern

    def _generate_palette(self) -> "Palette":
        """Creates a new palette with the dmc objects"""
        logger.info("Generating SVG palette...")
        image_palette = self._image.getpalette()

        palette = []
        for i in range(self._colors_num):
            r = i * 3
            red, green, blue = image_palette[r : r + 3]

            color = RGB(red, green, blue)
            matching_color = self._provider.get_matching_color(color)
            palette.append(matching_color)

        return palette

    def _clean_up(self, pattern: "PaletteImage") -> None:
        """Perform extra jobs like cleaning the image  removing isolated pixels."""
        logger.info("Cleaning up proces started...")

        for x in range(self.size.width):
            for y in range(self.size.height):
                coordinate = Coordinate(y, x)
                neighbours = get_neighbours(coordinate, pattern)

                if index_of(neighbours, pattern[y][x]) == -1:
                    mode = max(neighbours, key=neighbours.count)
                    pattern[y][x] = mode



import argparse
import importlib.metadata
import logging
import os

from tarraz import constants
from tarraz.logger import logger
from tarraz.processor import Tarraz
from tarraz.providers import DMCProvider
from tarraz.stitcher import DisplayStitcher, SVGStitcher
from tarraz.utils import color_choices, file_choices, parser

VERSION = importlib.metadata.version("tarraz")


def init_argparse() -> argparse.ArgumentParser:
    parser.add_argument(
        "--version",
        action="version",
        version=f"{parser.prog} version {VERSION}",
    )
    parser.add_argument(
        "image",
        type=lambda f: file_choices(constants.IMAGE_EXTENSIONS, f),
        help="Input image",
    )
    parser.add_argument(
        "-c",
        "--colors",
        type=int,
        default=3,
        help="Number of colors to use in the pattern.",
    )
    parser.add_argument(
        "-n",
        "--stitches-count",
        type=int,
        default=50,
        help="Number of stitches to use in the x axis.",
    )
    parser.add_argument(
        "-w",
        "--width",
        type=int,
        default=1000,
        help="Result pattern width.",
    )
    parser.add_argument(
        "-m",
        "--dmc",
        type=lambda f: file_choices(constants.COLORS_EXTENSIONS, f),
        help="DMC json color path.",
    )
    parser.add_argument(
        "-t",
        "--transparent",
        nargs="+",
        type=lambda f: color_choices(f),
        help="A Color to ignore from the end result.",
    )
    parser.add_argument(
        "-o",
        "--dist",
        type=str,
        default=constants.BASE_DIR / ".tmp/",
        help="Output destination directory.",
    )
    parser.add_argument(
        "-z",
        "--cell-size",
        type=int,
        default=10,
        help="The size of the generated Aida fabric cell.",
    )
    parser.add_argument(
        "--no-cleanup",
        action="store_true",
        help="Don't run cleanup job on generated image.",
    )
    parser.add_argument(
        "--svg",
        action="store_true",
        help="Export result to svg files.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Show debug messages.",
    )
    return parser


def main() -> None:
    p = init_argparse()
    args = p.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)

    base_file_name = os.path.basename(args.image).split(".")[0]

    logger.info("Generating pattern for %s...", base_file_name)
    logger.debug("\t File path: %s", args.image)
    logger.debug("\t X count: %s", args.stitches_count)
    logger.debug("\t Colors number: %s", args.colors)
    logger.debug("\t Result width: %s", args.width)
    logger.debug("\t DMC path: %s", args.dmc)
    logger.debug("\t No cleanup: %s", args.no_cleanup)
    logger.debug("\t SVG cell size: %s", args.cell_size)
    logger.debug("\t Destination: %s", args.dist)

    if args.transparent:
        logger.info("Transparent colors: %s", args.transparent)

    provider = DMCProvider(data_path=args.dmc)
    tarraz = Tarraz(
        args.image,
        provider=provider,
        x_count=args.stitches_count,
        colors_num=args.colors,
        result_width=args.width,
        cleanup=not args.no_cleanup,
    )

    pattern, colors = tarraz.process()

    if args.svg:
        SVGStitcher.stitch(
            pattern,
            colors,
            tarraz.size,
            transparent=args.transparent,
            configs=constants.SVG_VARIANTS,
            cell_size=args.cell_size,
            save_to=f"{args.dist}/{base_file_name}",
        )
    else:
        DisplayStitcher.stitch(
            pattern,
            colors,
            tarraz.size,
            transparent=args.transparent,
            cell_size=args.cell_size,
            save_to=f"{args.dist}/{base_file_name}",
        )

    logger.info("Tarraz process finished successfully!")


if __name__ == "__main__":
    exit(main())



import os
from abc import ABC
from typing import TYPE_CHECKING, List, Optional

from tarraz import constants
from tarraz.logger import logger
from tarraz.models import Coordinate

if TYPE_CHECKING:
    from tarraz.models import RGB, Color, ImageSize, Palette, PaletteImage


class Stitcher(ABC):
    def __init__(
        self,
        name: str,
        black_white: bool = False,
        minor_lines: bool = False,
        symbols: bool = True,
        scale: int = 20,
        save_to: "Optional[str]" = None,
        *args,
        **kwargs,
    ) -> None:
        self._black_white = black_white
        self._minor_lines = minor_lines
        self._symbols = symbols
        self._scale = scale

        self.name = name
        self.result = None

        self._dist_dir = save_to if save_to else constants.BASE_DIR / ".tmp"
        os.makedirs(self._dist_dir, exist_ok=True)

    def init(self, width: int, height: int) -> NotImplemented:
        return NotImplemented

    def finish(self) -> NotImplemented:
        return NotImplemented

    def save(self, ext: str) -> None:
        logger.debug("Create exports directory: %s...", self._dist_dir)

        file_name = f"{self._dist_dir}/{self.name}.{ext}"
        with open(file_name, "w") as f:
            logger.info(f"Saving {file_name}")
            f.write(self.result)

    def generate_key(
        self,
        colors: "Palette",
        size: int = 40,
        transparent: "Optional[List[RGB]]" = None,
    ) -> NotImplemented:
        return NotImplemented

    def draw_cell(
        self, coordinate: "Coordinate", size: int, color: "Optional[Color]" = None
    ) -> NotImplemented:
        return NotImplemented

    def draw_cells(
        self,
        pattern: "PaletteImage",
        colors: "Palette",
        cell_size: int,
        size: "ImageSize",
        config: "Optional[dict]" = None,
        transparent: "Optional[List[RGB]]" = None,
    ):
        if not transparent:
            transparent = []

        x = y = cell_size
        for row in pattern:
            for color_i in row:
                coordinate = Coordinate(x, y)
                color = (
                    colors[color_i] if colors[color_i].rgb not in transparent else None
                )
                self.draw_cell(coordinate, cell_size, color=color)
                x += cell_size
            y += cell_size
            x = cell_size

    @classmethod
    def stitch(
        cls,
        pattern: "PaletteImage",
        colors: "Palette",
        size: "ImageSize",
        cell_size: int = 10,
        configs: "Optional[dict]" = None,
        ext: "Optional[str]" = None,
        key_size: int = 40,
        save_to: "Optional[str]" = None,
        transparent: "Optional[List[RGB]]" = None,
    ):
        """Export picture with given variants.
        Supported variants:
          * Black/white.
          * Colored picture.
          * Colored with symbols.
        """
        logger.info("Stitching job started...")
        width = size.width * cell_size
        height = size.height * cell_size

        if not transparent:
            transparent = []

        if not configs:
            configs = [{"name": "NO_CONFIG"}]

        for config in configs:
            variant = cls(**config, save_to=save_to)

            if config.get("key"):
                variant.init(key_size * 13, key_size * len(colors))
                variant.generate_key(colors, key_size, transparent=transparent)
            else:
                variant.init(width, height)
                variant.draw_cells(
                    pattern,
                    colors,
                    cell_size,
                    size,
                    config=config,
                    transparent=transparent,
                )

            variant.finish()
            variant.save(ext)

    def __str__(self):
        return f"Stitcher<{self.name}>"



import json
from abc import ABC
from typing import List, Optional

from tarraz.logger import logger
from tarraz.models import Color, RGB
from tarraz.utils import euclidean_distance


class ColorProvider(ABC):
    def __init__(
        self,
        data_path: Optional[str] = None,
        colors: Optional[List["Color"]] = None,
    ) -> None:
        if not data_path and not colors:
            raise ValueError(
                "You need to supply either a data path or a colors list..."
            )

        self.matching_colors = {}
        self._data_path = data_path
        self.colors = colors if colors else self._read_colors()

        logger.debug("%d colors successfully processed.", len(self.colors))

    def _read_colors(self) -> "List[Color]":
        logger.info("Reading colors from %s", self._data_path)

        with open(self._data_path, mode="r") as data_file:
            colors = Color.create(json.load(data_file))

        return colors

    def get_matching_color(self, rgb_color: "RGB") -> "Optional[Color]":
        logger.debug("Getting a matching color for color %s.", rgb_color.css)

        if rgb_color in self.matching_colors:
            logger.debug("Loading matching color from cache...")
            return self.matching_colors[rgb_color]

        best_matching_index = self._get_best_color_index(rgb_color)

        if best_matching_index >= 0:
            matching_color = self.colors[best_matching_index]
            logger.debug(
                "Found %s as matching color for color %s.",
                matching_color.rgb.css,
                rgb_color.css,
            )
        else:
            matching_color = None
            logger.debug("No matching color found for %s color.", rgb_color.css)

        logger.debug("Storing matching color in cache...")
        self.matching_colors[rgb_color] = matching_color

        return matching_color

    def _get_best_color_index(self, rgb_color: "RGB") -> int:
        min_distance = float("inf")
        min_distance_index = float("-inf")

        for i, provider_color in enumerate(self.colors):
            dist = euclidean_distance(provider_color.rgb, rgb_color)

            if dist < min_distance:
                min_distance = dist
                min_distance_index = i

        return min_distance_index

    def index(self, rgb_color: "RGB") -> int:
        for i, c in enumerate(self.colors):
            if c.rgb == rgb_color:
                return i
        return -1

    def get(self, rgb_color: "RGB", default: "Optional" = None) -> "Optional[Color]":
        index = self.index(rgb_color)

        if index < 0:
            return default

        return self.colors[index]

    def __str__(self):
        name = self.__class__.__name__
        if "Provider" in name:
            name = name.split("Provider")[0]

        return f"{name}"



import os
from typing import TYPE_CHECKING, List, Optional

from tarraz.logger import logger
from tarraz.models import Color, Coordinate
from tarraz.stitcher import Stitcher

if TYPE_CHECKING:
    from tarraz.models import (
        RGB,
        ImageSize,
        Palette,
        PaletteImage,
        StrokeType,
        SVGAttributes,
    )


class SVGStitcher(Stitcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.result = ""
        self.file_name = f"{self._dist_dir}/{self.name}.svg"

        if os.path.exists(self.file_name):
            logger.debug(f"Removing {self.file_name}...")
            os.remove(self.file_name)

    def init(self, width: int, height: int) -> None:
        self._append_to_file(
            f"""
        <svg xmlns="http://www.w3.org/2000/svg"
            width="{width}"
            height="{height}"
            style ="fill:none;">
        """
        )
        self._append_to_file(
            """
            <style>
                .svg_txt    {
                    font-size:20px;
                }
                .glyph  {
                    stroke:#000000;
                    stroke-width:1;
                    stroke:1;
                }
            </style>
        """
        )

    def finish(self) -> None:
        self._append_to_file("</svg>")

    def generate_key(
        self,
        colors: "Palette",
        size: int = 40,
        transparent: "Optional[List[RGB]]" = None,
    ):
        keyed = set()
        if not transparent:
            transparent = []

        y = 0
        for i in range(len(colors)):
            if colors[i].rgb in transparent:
                continue

            color = colors[i]
            if color.code in keyed:
                continue

            self._add_key_to_svg(Coordinate(0, y), size, color)
            y += size
            keyed.add(color.code)

    def draw_cell(
        self, coordinate: "Coordinate", size: int, color: "Optional[Color]" = None
    ) -> None:
        fill, symbols, stroke = self._get_svg_attributes(coordinate, size, color=color)

        self._append_to_file(
            f"""
            <rect
                x="{coordinate.x}"
                y="{coordinate.y}"
                width="{size}"
                height="{size}"
                style="{fill}{stroke}"
            />
            {symbols}
        """
        )

    def draw_cells(
        self,
        pattern: "PaletteImage",
        palette: "Palette",
        cell_size: int,
        size: "ImageSize",
        config: "Optional[dict]" = None,
        transparent: "Optional[List[RGB]]" = None,
    ):
        super().draw_cells(
            pattern,
            palette,
            cell_size,
            size,
            config=config,
            transparent=transparent,
        )

        if config.get("mid_arrows"):
            self._mid_arrows(cell_size, size.width, size.height)

        if config.get("major_gridlines"):
            self._major_gridlines(cell_size, size.width, size.height)

    def _append_to_file(self, data: str) -> None:
        logger.debug("Appending data to file: %s...", self.file_name)

        with open(self.file_name, "a") as f:
            f.write(data)

    def save(self, ext: str) -> None:
        return

    def _major_gridlines(self, size: int, width: int, height: int) -> None:
        start = size + size * 10
        step = size * 10

        for x in range(start, width, step):
            self._append_to_file(
                f"""
                <line
                    x1="{x}"
                    y1="{size}"
                    x2="{x}"
                    y2="{height}"
                    style="stroke:black; stroke-width:2"
                />
            """
            )

        for y in range(start, height, step):
            self._append_to_file(
                f"""
                <line
                    x1="{size}"
                    y1="{y}"
                    x2="{width}"
                    y2="{y}"
                    style="stroke:black; stroke-width:2"
                />
            """
            )

    def _mid_arrows(self, size: int, width: int, height: int) -> None:
        h = size // 2

        self._append_to_file(
            f"""
            <path
                d="M0 {h}L{size} {h}M{h} 0L{size} {h} {h} {size}"
                stroke="black"
                stroke-width="2"
                fill="none"
                transform="translate(0 {height // 2})"
            />
        """
        )

        self._append_to_file(
            f"""
            <path
                d="M{h} 0L{h} {size} M{size} {h}L{h} {size} 0 {h}"
                stroke="black"
                stroke-width="2"
                fill="none"
                transform="translate({width // 2} 0)"
            />
        """
        )

    @staticmethod
    def _gen_glyph(color: "Color", coordinate: "Coordinate", scale: float = 1.0) -> str:
        transform = f"translate({coordinate.x} {coordinate.y}) scale({scale})"
        fill = ""

        if not color.glyph:
            color.assign_glyph()

        if color.glyph == 0:
            # Backslash
            path = "M4 4L16 16"
        elif color.glyph == 1:
            # Forward slash
            path = "M4 16L16 4M4 10L 16 10"
        elif color.glyph == 2:
            # Black little square
            path = "M7 7L7 13 13 13 13 7Z"
            fill = "black"
        elif color.glyph == 3:
            #
            path = "M4 4L10 16L16 4 Z"
        elif color.glyph == 4:
            # Diagonal cross
            path = "M4 4L16 16M4 16 L16 4"
        elif color.glyph == 5:
            # Square
            path = "M4 4L4 16 16 16 16 4Z"
        elif color.glyph == 6:
            # Upside down black triangle
            path = "M4 4L10 16L16 4 Z"
            fill = "black"
        elif color.glyph == 7:
            # Black diamond
            path = "M10 4L6 10 10 16 14 10Z"
            fill = "black"
        elif color.glyph == 8:
            # Little square
            path = "M8 8L8 12 12 12 12 8Z"
        elif color.glyph == 9:
            # 8-way cross
            path = "M4 4L16 16M4 16 L16 4M10 4L10 16M4 10L16 10"
        elif color.glyph == 10:
            # Black Square
            path = "M4 4L4 16 16 16 16 4Z"
            fill = "black"
        else:
            return ""

        return f"""
            <path
                class="glyph"
                d="{path}"
                fill="{fill}"
                transform="{transform}"
            />
        """

    def _add_key_to_svg(
        self, coordinate: "Coordinate", size: int, color: "Color"
    ) -> None:
        fill, symbols, stroke = self._get_svg_attributes(coordinate, size, color)

        self._append_to_file(
            f"""
            <rect
                x="0"
                y="{coordinate.y}"
                width="{size}"
                height="{size}"
                style="{fill}{stroke}"
            />
            {symbols}
        """
        )

        # Color name
        self._append_to_file(
            f"""
            <rect
                x="{size}"
                y="{coordinate.y}"
                width="{size * 10}"
                height="{size}"
                style="fill: rgb(255,255,255); stroke: black; stroke-width: 1;"
            />
        """
        )
        self._append_to_file(
            f"""
            <text
                x="{coordinate.x + size * 1.5}"
                y="{coordinate.y + size / 2}"
                fill="black"
            >
                {color.name}
            </text>
        """
        )

        # Color code
        self._append_to_file(
            f"""
            <rect
                x="{size * 11}"
                y="{coordinate.y}"
                width="{size * 2}"
                height="{size}"
                style="fill: rgb(255,255,255); stroke: black; stroke-width: 1;"
            />
        """
        )
        self._append_to_file(
            f"""
            <text
                x="{size * 11 + (size / 2)}"
                y="{coordinate.y + size / 2}"
                fill="black"
            >
                {color.code}
            </text>
        """
        )

    def _get_svg_attributes(
        self, coordinate: "Coordinate", size: int, color: "Optional[Color]" = None
    ) -> "SVGAttributes":
        fill = "fill:rgb(255,255,255);"
        stroke: "StrokeType" = "stroke:none;"
        symbols = ""

        scale = size / self._scale

        if not color:
            fill = f"fill:transparent;"
        elif not self._black_white:
            fill = f"fill:{color.rgb.css};"

        if self._minor_lines:
            stroke = "stroke:rgb(20,20,20);stroke-width:1;"

        if color and self._symbols:
            symbols = self._gen_glyph(color, coordinate, scale)

        return fill, symbols, stroke



import argparse
import bisect
import math
import os
from bisect import bisect_left
from typing import TYPE_CHECKING, List

from PIL import Image, ImageColor

from tarraz.logger import logger
from tarraz.models import ImageSize, RGB

if TYPE_CHECKING:
    from PIL.Image import Image as ImageType
    from tarraz.models import Coordinate, RGBImage


parser = argparse.ArgumentParser(
    description="Generate a DMC-colored cross-stitch pattern from a given image.",
)


def color_choices(value: str) -> "RGB":
    value = value.strip()

    try:
        value = (
            ImageColor.getrgb(value)
            if value.startswith("#")
            else [int(c) for c in value.split(",")]
        )
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid color value '{value}'")

    return RGB(*value)


def file_choices(choices: List[str], file_name: str) -> str:
    _, ext = os.path.splitext(file_name)
    if ext.upper() not in map(str.upper, choices):
        raise argparse.ArgumentTypeError(
            f"Unsupported file format '{ext}'. Available formats are: {choices}."
        )
    return file_name


def index_of(iterable, element, sort=False) -> int:
    arr = sorted(iterable) if sort else iterable

    i = bisect_left(arr, element)
    if i != len(arr) and arr[i] == element:
        return i

    return -1


def euclidean_distance(rgb_1: "RGB", rgb_2: "RGB") -> float:
    red_point = (rgb_1.red - rgb_2.red) ** 2
    green_point = (rgb_1.green - rgb_2.green) ** 2
    blue_point = (rgb_1.blue - rgb_2.blue) ** 2

    return math.sqrt(red_point + green_point + blue_point)


def generate_image(data: "RGBImage") -> "ImageType":
    """Create a resized image from a given matrix of colors."""
    size = ImageSize(len(data[0]), len(data))
    logger.info(
        f"Generating a new image [{size.width}x{size.height}] for the new colors..."
    )

    image = Image.new("RGB", size)

    data = [rgb.native for row in data for rgb in row]
    image.putdata(data)

    return image


def get_neighbours(coordinate: "Coordinate", matrix: List[List[int]]) -> List[int]:
    x, y = coordinate
    rows = len(matrix)
    cols = len(matrix[0]) if rows else 0

    width = 1
    neighbours: List[int] = []

    for i in range(max(0, x - width), min(rows, x + width + 1)):
        for j in range(max(0, y - width), min(cols, y + width + 1)):
            if not (i == x and j == y):
                # Insort neighbors sorted
                bisect.insort_left(neighbours, matrix[i][j])

    return neighbours



from typing import List, Optional

from tarraz.colors import DMC_COLORS
from tarraz.models import Color
from tarraz.providers import ColorProvider


class DMCProvider(ColorProvider):
    def __init__(
        self,
        data_path: Optional[str] = None,
        colors: Optional[List["Color"]] = None,
    ) -> None:
        if not data_path and not colors:
            colors = Color.create(DMC_COLORS)

        super().__init__(data_path=data_path, colors=colors)
